import unittest


class ExternalTest(unittest.TestCase):

    def test_category_encoders(self):
        try:
            import category_encoders  # noqa F401
        except ImportError:
            pass
        else:
            from ballet.eng.category_encoders import BackwardDifferenceEncoder  # noqa F401

    def test_feature_engine(self):
        try:
            import feature_engine  # noqa F401
        except ImportError:
            pass
        else:
            from ballet.eng.feature_engine import CountFrequencyCategoricalEncoder  # noqa F401

    def test_featuretools(self):
        try:
            import featuretools  # noqa F401
        except ImportError:
            pass
        else:
            from ballet.eng.featuretools import DFSTransformer  # noqa F401

    def test_skits(self):
        try:
            import skits  # noqa F401
        except ImportError:
            pass
        else:
            from ballet.eng.skits import AutoregressiveTransformer  # noqa F401

    def test_sklearn(self):
        from ballet.eng.sklearn import KNNImputer  # noqa F401

    def test_tsfresh(self):
        try:
            import tsfresh  # noqa F401
        except ImportError:
            pass
        else:
            from ballet.eng.tsfresh import FeatureAugmenter  # noqa F401



from ballet.validation.common import subsample_data_for_validation


def validate_feature_acceptance(accepter_class, feature, features, X_df, y,
                                subsample):
    if subsample:
        X_df, y = subsample_data_for_validation(X_df, y)
    accepter = accepter_class(X_df, y, features, feature)
    return accepter.judge()



try:
    from feature_engine.categorical_encoders import CountFrequencyCategoricalEncoder
    from feature_engine.categorical_encoders import MeanCategoricalEncoder
    from feature_engine.categorical_encoders import OneHotCategoricalEncoder
    from feature_engine.categorical_encoders import OrdinalCategoricalEncoder
    from feature_engine.categorical_encoders import RareLabelCategoricalEncoder
    from feature_engine.categorical_encoders import WoERatioCategoricalEncoder
    from feature_engine.discretisers import DecisionTreeDiscretiser
    from feature_engine.discretisers import EqualFrequencyDiscretiser
    from feature_engine.discretisers import EqualWidthDiscretiser
    from feature_engine.discretisers import UserInputDiscretiser
    from feature_engine.missing_data_imputers import AddMissingIndicator
    from feature_engine.missing_data_imputers import ArbitraryNumberImputer
    from feature_engine.missing_data_imputers import CategoricalVariableImputer
    from feature_engine.missing_data_imputers import EndTailImputer
    from feature_engine.missing_data_imputers import MeanMedianImputer
    from feature_engine.missing_data_imputers import RandomSampleImputer
    from feature_engine.outlier_removers import ArbitraryOutlierCapper
    from feature_engine.outlier_removers import OutlierTrimmer
    from feature_engine.outlier_removers import Winsorizer
    from feature_engine.variable_transformers import BoxCoxTransformer
    from feature_engine.variable_transformers import LogTransformer
    from feature_engine.variable_transformers import PowerTransformer
    from feature_engine.variable_transformers import ReciprocalTransformer
    from feature_engine.variable_transformers import YeoJohnsonTransformer
except ImportError:
    pass


__all__ = (
    'AddMissingIndicator',
    'ArbitraryNumberImputer',
    'ArbitraryOutlierCapper',
    'BoxCoxTransformer',
    'CategoricalVariableImputer',
    'CountFrequencyCategoricalEncoder',
    'DecisionTreeDiscretiser',
    'EndTailImputer',
    'EqualFrequencyDiscretiser',
    'EqualWidthDiscretiser',
    'LogTransformer',
    'MeanCategoricalEncoder',
    'MeanMedianImputer',
    'OneHotCategoricalEncoder',
    'OrdinalCategoricalEncoder',
    'OutlierTrimmer',
    'PowerTransformer',
    'RandomSampleImputer',
    'RareLabelCategoricalEncoder',
    'ReciprocalTransformer',
    'UserInputDiscretiser',
    'Winsorizer',
    'WoERatioCategoricalEncoder',
    'YeoJohnsonTransformer',
)



import random

from ballet.util.log import logger
from ballet.util.testing import seeded
from ballet.validation.base import FeatureAcceptanceMixin, FeatureAccepter
from ballet.validation.common import RandomFeaturePerformanceEvaluator
from ballet.validation.entropy import estimate_conditional_information
from ballet.validation.gfssf import (
    GFSSFIterationInfo, GFSSFPerformanceEvaluator, _compute_lmbdas,
    _compute_threshold, _concat_datasets)


class NeverAccepter(FeatureAccepter):

    def judge(self):
        logger.info('Judging feature using {}'.format(self))
        return False


class RandomAccepter(FeatureAcceptanceMixin,
                     RandomFeaturePerformanceEvaluator):

    def judge(self):
        """Accept feature with probability p"""
        logger.info('Judging feature using {}'.format(self))
        with seeded(self.seed):
            return random.uniform(0, 1) < self.p


class AlwaysAccepter(FeatureAccepter):
    def judge(self):
        logger.info('Judging feature using {}'.format(self))
        return True


class GFSSFAccepter(FeatureAcceptanceMixin, GFSSFPerformanceEvaluator):

    def judge(self):
        """Judge feature acceptance using GFSSF

        Uses lines 1-8 of agGFSSF where we do not remove accepted but
        redundant features on line 8.
        """

        logger.info(f'Judging feature using {self}')

        feature_df_map = self._get_feature_df_map()

        candidate_df = feature_df_map[self.candidate_feature]
        n_samples, n_candidate_cols = candidate_df.shape

        lmbda_1, lmbda_2 = _compute_lmbdas(
            self.lmbda_1, self.lmbda_2, feature_df_map)

        logger.debug(
            f'Recomputed lambda_1={lmbda_1:0.4f}, lambda_2={lmbda_2:0.4f}')

        info = []

        omit_in_test = [None, *self.features]
        n_omit = len(omit_in_test)
        for i, omitted_feature in enumerate(omit_in_test):

            z = _concat_datasets(
                feature_df_map, n_samples,
                omit=[self.candidate_feature, omitted_feature])

            # Calculate CMI of candidate feature
            cmi = estimate_conditional_information(candidate_df, self.y, z)

            if omitted_feature is not None:
                omit_df = feature_df_map[omitted_feature]
                _, n_omit_cols = omit_df.shape

                # Calculate CMI of omitted feature
                cmi_omit = estimate_conditional_information(
                    omit_df, self.y, z)
            else:
                cmi_omit = 0
                n_omit_cols = 0

                # want to log to INFO only the case of I(Z|Y;X) where X is the
                # entire feature matrix, i.e. no omitted features.
                logger.info(f'I(feature ; target | existing_features) = {cmi}')

            statistic = cmi - cmi_omit
            threshold = _compute_threshold(
                lmbda_1, lmbda_2, n_candidate_cols, n_omit_cols)
            delta = statistic - threshold

            if delta >= 0:
                omitted_source = getattr(omitted_feature, 'source', 'None')
                logger.debug(
                    f'Succeeded while omitting feature: {omitted_source}')

                return True
            else:
                iteration_info = GFSSFIterationInfo(
                    i=i,
                    n_samples=n_samples,
                    candidate_feature=self.candidate_feature,
                    candidate_cols=n_candidate_cols,
                    candidate_cmi=cmi,
                    omitted_feature=omitted_feature,
                    omitted_cols=n_omit_cols,
                    omitted_cmi=cmi_omit,
                    statistic=statistic,
                    threshold=threshold,
                    delta=delta,
                )
                info.append(iteration_info)
                logger.debug(
                    f'Completed iteration {i}/{n_omit}: {iteration_info}')

        info_closest = max(info, key=lambda x: x.delta)
        cmi_closest = info_closest.candidate_cmi
        omitted_cmi_closest = info_closest.omitted_cmi
        statistic_closest = info_closest.statistic
        threshold_closest = info_closest.threshold
        logger.info(
            f'Rejected feature: best marginal conditional mutual information was not greater than threshold ({cmi_closest:0.4f} - {omitted_cmi_closest:0.4f} = {statistic_closest:0.4f}, vs needed {threshold_closest:0.4f}).')  # noqa

        return False



try:
    from skits.feature_extraction import AutoregressiveTransformer
    from skits.feature_extraction import SeasonalTransformer
    from skits.feature_extraction import IntegratedTransformer
    from skits.feature_extraction import RollingMeanTransformer
    from skits.feature_extraction import TrendTransformer
    from skits.feature_extraction import FourierTransformer
    from skits.preprocessing import ReversibleImputer
    from skits.preprocessing import DifferenceTransformer
    from skits.preprocessing import LogTransformer
    from skits.preprocessing import HorizonTransformer
except ImportError:
    pass


__all__ = (
    'AutoregressiveTransformer',
    'DifferenceTransformer',
    'FourierTransformer',
    'HorizonTransformer',
    'IntegratedTransformer',
    'LogTransformer',
    'ReversibleImputer',
    'RollingMeanTransformer',
    'SeasonalTransformer',
    'TrendTransformer',
)



try:
    from category_encoders import BackwardDifferenceEncoder
    from category_encoders import BaseNEncoder
    from category_encoders import BinaryEncoder
    from category_encoders import CatBoostEncoder
    from category_encoders import CountEncoder
    from category_encoders import GLMMEncoder
    from category_encoders import HashingEncoder
    from category_encoders import HelmertEncoder
    from category_encoders import JamesSteinEncoder
    from category_encoders import LeaveOneOutEncoder
    from category_encoders import MEstimateEncoder
    from category_encoders import OneHotEncoder
    from category_encoders import OrdinalEncoder
    from category_encoders import PolynomialEncoder
    from category_encoders import SumEncoder
    from category_encoders import TargetEncoder
    from category_encoders import WOEEncoder
except ImportError:
    pass


__all__ = (
    'BackwardDifferenceEncoder',
    'BaseNEncoder',
    'BinaryEncoder',
    'CatBoostEncoder',
    'CountEncoder',
    'GLMMEncoder',
    'HashingEncoder',
    'HelmertEncoder',
    'JamesSteinEncoder',
    'LeaveOneOutEncoder',
    'MEstimateEncoder',
    'OneHotEncoder',
    'OrdinalEncoder',
    'PolynomialEncoder',
    'SumEncoder',
    'TargetEncoder',
    'WOEEncoder',
)
